/*
 + DISCLAIMER
  USE AT YOUR OWN RISK.

 + TECHNICAL DESCRIPTION
  ɉxg񂳂Ɓcc
  [] = crypted, {} = compressed
  <Header>
  Algo: TH_CRYPTER
  TH_CRYPTER param: key=1Bh step=37h block=10h limit=10h
  [
    +00 <magic "THA1">.4b
    +04 <original size of entries +123456789>.d
    +08 <compressed size of entries +987654321>.d 
    +0c <number of entry-F7E7F8ACh>.d
  ]

  <Data>
  Algo: TH_LZ -> TH_CRYPTER
  TH_CRYPTER param: crypt_param[th_get_cryprm_index(<file name>)]
  { [ +00 <data>.?b ] }

  <File entries>
  Algo: TH_LZ -> TH_CRYPTER
  TH_CRYPTER param: key=3Eh step=9Bh block=80h limit=same as compressed entries size
  {
    [ (
      +00 <file name>.?b (NULL terminated, varibale length (in multiples of 4))
      +xx <offset>.l
      +yy <original size>.l
      ) * <number of entry>
    ]
  }
*/

#include <stdio.h>
#include <stdlib.h>

void th_crypter(unsigned char* in, unsigned char *out, int size, unsigned char key, unsigned char step, int block, int limit)
{
	unsigned char *i_itr = in;
	unsigned char *o_itr = out;
	unsigned char *o_base;
	int addup;
	int i;

	addup = size % block;
	if(addup >= block / 4) addup = 0;
	addup += size % 2;
	size -= addup;

	while(size > 0 && limit > 0) {
		if(size < block) block = size;
		/* first phase */
		o_base = o_itr;
		o_itr = o_itr + block - 1;
		for(i = (block + 1) / 2; i > 0; --i, ++i_itr) {
			/* decode char */
			*o_itr = *i_itr ^ key;
			o_itr -= 2;
			/* update key */
			key += step;
		}
		/* second phase */
		o_itr = o_base + block - 2;
		for(i = block / 2; i > 0; --i, ++i_itr) {
			/* decode char */
			*o_itr = *i_itr ^ key;
			o_itr -= 2;
			/* update key */
			key += step;
		}
		/* seek pointers */
		size -= block;
		o_itr = o_base + block;
		limit -= block;
	}
	/* transfer remaining data */
	size += addup;
	if(size > 0) memcpy(o_itr, i_itr, size);
}

struct bitstrm {
	unsigned char b;
	unsigned char c;
	unsigned char *cur;
	unsigned char *base;
	int size;
};

void bitstrm_init(struct bitstrm *strm, unsigned char *buf, int size)
{
	strm->b = 0x80;
	strm->c = 0;
	strm->cur = strm->base = buf;
	strm->size = size;
}

int bitstrm_read(struct bitstrm *strm, int bits)
{
	int ret = 0;
	int i;

	for(i = 1 << (bits - 1); i > 0; i >>= 1) {
		if(strm->b == 0x80) {
			if(strm->cur - strm->base < strm->size)
				strm->c = *strm->cur++;
			else
				strm->c = 0;
		}
		if(strm->b & strm->c) ret |= i;
		strm->b >>= 1;
		if (!strm->b) strm->b = 0x80;
	}
	return ret;
}

/* TH08 LZ (18bit LZSS, Dictionary = 8KB)
 * 1.b <char>.B
 * 0.b <offset>.13b <length>.4b
 * 0.b 0.13b (EOF) */
#define DICT_SIZE 0x2000
void th_unlz(unsigned char *in, int insize, unsigned char *out, int outsize)
{
	unsigned char g_dict[DICT_SIZE];
	struct bitstrm bs;
	int dicofs = 1;
	int patofs;
	int patlen;
	int c, i;

	memset(g_dict, 0, DICT_SIZE);
	bitstrm_init(&bs, in, insize);

	while(outsize > 0) {
		/* flag (1 bit) */
		int flag = bitstrm_read(&bs, 1);
		if(flag) {	/* plain */
			/* plain char (8 bits) */
			c = bitstrm_read(&bs, 8);
			*out++ = c;
			g_dict[dicofs] = c;
			dicofs = (dicofs + 1) % DICT_SIZE;
			--outsize;
		} else {	/* compressed */
			/* offset to the pattern (13 bits) */
			patofs = bitstrm_read(&bs, 13);
			if(!patofs) break;
			/* length of the pattern - 2 (4 bits) */
			patlen = bitstrm_read(&bs, 4) + 2;
			/* decompress data */
			for(i = 0; i <= patlen && outsize > 0; ++i, --outsize) {
				c = g_dict[(patofs + i) % DICT_SIZE];
				*out++ = c;
				g_dict[dicofs] = c;
				dicofs = (dicofs + 1) % DICT_SIZE;
			}
		}
	}
}

struct {
	unsigned char key;
	unsigned char step;
	int block;
	int limit;
} g_cryprm[] = {
	{ 0x1b, 0x37, 0x40,  0x2800 },
	{ 0x51, 0xe9, 0x40,  0x3000 },
	{ 0xc1, 0x51, 0x80,  0x3200 },
	{ 0x03, 0x19, 0x400, 0x7800 },
	{ 0xab, 0xcd, 0x200, 0x2800 },
	{ 0x12, 0x34, 0x80,  0x3200 },
	{ 0x35, 0x97, 0x80,  0x2800 },
	{ 0x99, 0x37, 0x400, 0x2000 }
};

struct th_file_entry {
	char *name;
	int offset;
	int complen;
	int origlen;
};

#define ARRAY_SIZE(a) (sizeof(a) / sizeof(a[0]))

int th_get_cryprm_index(char *name)
{
	char index = 0;
	while(*name) index += *name++;
	return index & 7;
}

int th_extract_file(FILE *dat, struct th_file_entry *entry)
{
	FILE *fp;
	int i = th_get_cryprm_index(entry->name);
	char *cryptbuf, *compbuf, *outbuf;

	/* allocate buffers for extracting data */
	cryptbuf = malloc(entry->complen);
	if(!cryptbuf) return 0; 
	compbuf = malloc(entry->complen);
	if(!compbuf) {
		free(cryptbuf);
		return 0;
	}
	
	fseek(dat, entry->offset, SEEK_SET);
	fread(cryptbuf, 1, entry->complen, dat);

	th_crypter(cryptbuf, compbuf, entry->complen, g_cryprm[i].key, g_cryprm[i].step, g_cryprm[i].block, g_cryprm[i].limit);
	free(cryptbuf);
	
	if (entry->complen == entry->origlen) {
		outbuf = compbuf;
	} else {
		outbuf = malloc(entry->origlen);
		if(!outbuf) {
			free(compbuf);
			return 0;
		}
		
		th_unlz(compbuf, entry->complen, outbuf, entry->origlen);
		free(compbuf);
	}

	fp = fopen(entry->name, "wb");
	fwrite(outbuf, entry->origlen, 1, fp);
	fclose(fp);
	free(outbuf);
}

int main(int argc, char *argv[])
{
	FILE *fp;
	size_t file_size;
	int i;
	char hedbuf[16];
	char *cryptdir;
	char *compdir;
	char *dir, *entry_raw;
	struct th_file_entry *entries;
	struct {
		unsigned long magic;
		unsigned long dir_size;
		unsigned long dir_compsize;
		unsigned long unk2;
	} hed;

	if(argc < 3) {
		printf(
		"usage: thamlt <arch> <cmd> [<param> ...]\n"
		"Uncompress Touhou Bunkachou/Fuujinroku THA1 arch\n\n"
		"cmd:\n"
		"  l\tlist files\n"
		"  e\textract <param>\n"
		"  a\textract all\n");
		exit(EXIT_SUCCESS);
	}
	fp = fopen(argv[1], "rb");
	if(!fp) {
		perror("fopen");
		exit(EXIT_FAILURE);
	}
	/* read and decrypt header */
	fread(&hedbuf, 0x10, 1, fp);
	th_crypter(hedbuf, (char*)&hed, 0x10, 0x1B, 0x37, 0x10, 0x10);
	if(hed.magic != '1AHT') {
		fprintf(stderr, "th_read_header: inavlid file\n");
		exit(EXIT_FAILURE);
	}
	hed.dir_size     -= 123456789;
	hed.dir_compsize -= 987654321;
	hed.unk2         += 0xF7E7F8AC;

	/* read and decrypt file entries */
	fseek(fp, 0, SEEK_END);
	file_size = ftell(fp);
	rewind(fp);
	
	fseek(fp, file_size - hed.dir_compsize, SEEK_SET);
	cryptdir = malloc(hed.dir_compsize);
	fread(cryptdir, hed.dir_compsize, 1, fp);
	
	compdir = malloc(hed.dir_compsize);
	th_crypter(cryptdir, compdir, hed.dir_compsize, 0x3E, 0x9B, 0x80, hed.dir_size);
	free(cryptdir);
	
	dir = malloc(hed.dir_size);
	th_unlz(compdir, hed.dir_compsize, dir, hed.dir_size);
	free(compdir);
	
	entry_raw = dir;
	entries = malloc(hed.unk2 * sizeof(struct th_file_entry));
	for(i = 0; i < hed.unk2; ++i) {
		int ofs = 0;
		while(entry_raw[3 + ofs]) ofs += 4;
		entries[i].name    = strdup(entry_raw);
		entries[i].offset  = *(unsigned long*)&entry_raw[4 + ofs];
		entries[i].origlen = *(unsigned long*)&entry_raw[8 + ofs];
		entry_raw += 16 + ofs;
	}
	free(dir);

	for(i = 0; i < hed.unk2 - 1; ++i) {
		entries[i].complen = entries[i + 1].offset - entries[i].offset;
		if(entries[i].complen <= 0) {
			exit(EXIT_FAILURE);
		}
	}
	entries[i].complen = (file_size - hed.dir_compsize) - entries[i].offset;
	if(entries[i].origlen <= 0) {
		exit(EXIT_FAILURE);
	}
	
	/* extract files */
	switch(argv[2][0]) {
	case 'l':
		printf("%-16s%16s%8s\n", "Filename", "Filesize", "Ratio");
		printf("----------------------------------------\n");
		for(i = 0; i < hed.unk2; ++i) {
			printf("%-16s%#16d%#8d\n", entries[i].name, entries[i].origlen, entries[i].complen * 100 / entries[i].origlen);
			free(entries[i].name);
		}
		break;
	case 'a':
		printf("start extracting\n");
		for(i = 0; i < hed.unk2; ++i) {
			printf("Extracting %s...", entries[i].name);
			if(th_extract_file(fp, &entries[i]))
				printf("succeeded\n");
			else
				printf("failed\n");
			free(entries[i].name);
		}
		break;
	case 'e':
		if(argc < 4) {
			fprintf(stderr, "Error: not enough parameters\n");
			fclose(fp);
			free(entries);
			exit(EXIT_FAILURE);
		}
		for(i = 0; i < hed.unk2; ++i) {
			if(!strcmp(entries[i].name, argv[3])) {
				printf("Extracting %s...", entries[i].name);
				if(th_extract_file(fp, &entries[i]))
					printf("succeeded\n");
				else
					printf("failed\n");
			}
			free(entries[i].name);
		}
		break;
	}

	fclose(fp);
	free(entries);
	
	exit(EXIT_SUCCESS);
}
